package de.klawitter.android.widget;

import java.util.Arrays;
import java.util.BitSet;

import android.content.Context;
import android.util.AttributeSet;
import android.view.View;
import android.view.ViewGroup;
import android.widget.LinearLayout;

/**
 * Variant of Tablelayout, allowing Cells to span over multiple
 * rows.
 * <ul>
 * <li>The attributes "rows" and "columns" specify the size of the
 * Table.
 * <li>Children Views are being inserted on a row by row and column
 * by columns base.
 * <li>fill_parent will be used as an equivalent to "until the end
 * of the row/column". "wrap_content" means one row/column.
 * <li>If a child view implements Griddable, the attributes gridX,
 * gridY, gridWidth and gridHeight are being used instead.
 * <li>The table does not need to be completely filled.
 * <li>All columns will have the same width and all rows will have the
 * same height.
 * </ul>
 * @author <a href="mailto:info-android@klawitter.de">Holger Klawitter</a>
 */
public class GridLayout extends LinearLayout {

	public final static String SCHEMA = "http://schemas.klawitter.de/android/grid";

	private int rows;
	private int cols;

	public GridLayout(Context context, AttributeSet attrs) {
		super(context, attrs);
		cols = attrs.getAttributeIntValue(SCHEMA, "columns", 1);
		rows = attrs.getAttributeIntValue(SCHEMA, "rows", 1);
	}

	private LocationInfo[] viewInfo;
	// I assume the the order of children returned from getChildAt() to be stable
	// if not, I have to resort to something like this:
	//private HashMap<View,LocationInfo> viewInfo;


	private int[][] occupied;

	private static class LocationInfo {
		int x; int y; int w; int h; // position of the cell
		int mx; int my; // measured size
		public LocationInfo( int x, int y, int w, int h ) {
			this.x = x; this.y = y; this.w = w; this.h = h;
		}
		public String toString() {
			return "LocationInfo["+x+"x"+y+"@"+w+"x"+h+"]";
		}
	} // class LocationInfo

	/**
	 * Generates a readable error message.
	 */
	private void error( String msg, int cx, int cy, int w, int h, int idx ) {
		throw new IllegalArgumentException( msg + ": view #"+idx+" ("+cx+","+cy+"@"+w+","+h+")" );
	}

	/**
	 * Marks the area as being occupied by the View with the index idx.
	 * @param cx origin x (0 based)
	 * @param cy origin y (0 based)
	 * @param w number of columns occupied by this view
	 * @param h number of rows occupied by the view
	 * @param idx child index of this view
	 * @throws IllegalArgumentException, if the cell does not fit into the grid.
	 */
	private void fill( int cx, int cy, int w, int h, int idx ) throws IllegalArgumentException {
		if( cx<0 || cy<0 || cx>=cols || cy>=rows ) error( "cell out of bounds",cx,cy,w,h,idx);
		if( cx+w > cols ) error( "cell too wide ",cx,cy,w,h,idx);
		if( cy+h > rows ) error( "cell too high ",cx,cy,w,h,idx);
		for( int ii=0; ii<w; ii++ ) {
			for( int jj=0; jj<h; jj++ ) {
				if( occupied[cx+ii][cy+jj]!=FREE ) {
					error( "overlap with #"+occupied[cx+ii][cy+jj],cx,cy,w,h,idx );
				} else {
					occupied[cx+ii][cy+jj] = idx;
				}
			}
		}

		LocationInfo li = new LocationInfo(cx,cy,w,h); 
		viewInfo[idx] = li;	//viewInfo.put( c, li );
	}

	/** marker for unoccupied cells */
	private final static int FREE = -1;

	/**
	 * Retrieves the information from the child views and computes
	 * the layoutinformation for later use.
	 */
	private void setupGrid() {
		if( viewInfo!=null ) return;

		occupied = new int[cols][];
		for( int i=0; i<cols; i++ ) {
			occupied[i] = new int[rows]; 
			Arrays.fill(occupied[i], FREE );
		}
		assert occupied[0][0]==FREE;

		int len = getChildCount();
		int cx = 0;
		int cy = 0;
		//viewInfo = new HashMap<View,LocationInfo>();
		viewInfo = new LocationInfo[len];

		for( int i=0; i<len; i++ ) {
			if( cy>=rows ) {
				throw new IllegalArgumentException("grid full");
			}

			View c = getChildAt(i);
			int w = 1;
			int h = 1;
			int px = cx;
			int py = cy;

			if( c instanceof Griddable ) {
				GridLocation gl = ((Griddable) c).getGridLocation();
				if( gl.getGridY()!=GridLocation.UNDEFINED ) py = gl.getGridY();
				if( gl.getGridX()!=GridLocation.UNDEFINED ) px = gl.getGridX();
				if( gl.getGridWidth()!=GridLocation.UNDEFINED ) w = gl.getGridWidth();
				if( gl.getGridHeight()!=GridLocation.UNDEFINED ) h = gl.getGridHeight();
			} else {
				ViewGroup.LayoutParams lp = c.getLayoutParams();
				if( lp.height==LayoutParams.FILL_PARENT ) {
					h=rows-cy;
				} else {
					h=1;
				}
				if( lp.width==LayoutParams.FILL_PARENT ) {
					w=cols-cx;
				} else {
					w=1;
				}
			}

			fill( px, py, w, h, i );

			// advance to next free position
			while( cy<rows && occupied[cx][cy]!=FREE ) {
				cx++;
				if( cx>=cols ) {
					cx=0;
					cy++;
				}
			}
		}
	}

	private int computeRowWidth( int row ) {
		BitSet used = new BitSet();
		int w = 0;
		for( int i=0; i<cols; i++ ) {
			int idx = occupied[i][row];
			if( idx!=FREE && !used.get(idx) ) {
				w += viewInfo[idx].mx;
				used.set(idx);
			}
		}
		return w;
	}

	private int computeColHeight( int col ) {
		BitSet used = new BitSet();
		int h = 0;
		for( int i=0; i<rows; i++ ) {
			int idx = occupied[col][i];
			if( idx!=FREE && !used.get(idx) ) {
				h += viewInfo[idx].my;
				used.set(idx);
			}
		}
		return h;
	}

	private int computeMinWidth() {
		int min = 0;
		for( int i=0; i<rows; i++ ) { min = Math.max( min, computeRowWidth(i) ); }
		return min;
	}

	private int computeMinHeight() {
		int min = 0;
		for( int i=0; i<cols; i++ ) { min = Math.max( min, computeColHeight(i) ); }
		return min;
	}

	@Override 
	protected void onMeasure( int wSpec, int hSpec ) {
		setupGrid();
		int w = MeasureSpec.getSize(wSpec);
		int h = MeasureSpec.getSize(hSpec);
		int cellwidth = w/cols;
		int cellheight = h/rows;

		int len = getChildCount();
		for( int i=0; i<len; i++ ) {
			View c = getChildAt(i);
			LocationInfo li = viewInfo[i]; //viewInfo.get(c);
			assert li!=null;
			c.measure(
					resolveSize(li.w*cellwidth, MeasureSpec.UNSPECIFIED ),
					resolveSize(li.h*cellheight, MeasureSpec.UNSPECIFIED )
			);
			li.mx = c.getMeasuredWidth();
			li.my = c.getMeasuredHeight();
		}
		if( MeasureSpec.getMode(wSpec)==MeasureSpec.AT_MOST ) {
			w = Math.min( w, computeMinWidth() );
		}
		if( MeasureSpec.getMode(hSpec)==MeasureSpec.AT_MOST ) {
			h = Math.min( h, computeMinHeight() );
		}

		for( int i=0; i<len; i++ ) {
			View c = getChildAt(i);
			LocationInfo li = viewInfo[i]; //viewInfo.get(c);
			assert li!=null;
			c.measure(
					resolveSize(li.w*cellwidth, MeasureSpec.EXACTLY ),
					resolveSize(li.h*cellheight, MeasureSpec.EXACTLY )
			);
		}
		setMeasuredDimension( w, h );
	}

	@Override
	protected void onLayout( boolean changed, int l, int t, int r, int b ) {
		assert viewInfo!=null;
		int w = (r-l)/cols;
		int h = (b-t)/rows;
		int len = getChildCount();
		for( int i=0; i<len; i++ ) {
			View c = getChildAt(i);
			LocationInfo li = viewInfo[i]; //viewInfo.get(c);
			assert li!=null;
			c.layout( li.x * w, li.y * h, (li.x+li.w)*w, (li.y+li.h)*h );
		}		
	}

}
